from __future__ import annotations

AQUALINK_API_KEY = "EOOEMOW4YR6QNB07"

AQUALINK_LOGIN_URL = "https://prod.zodiac-io.com/users/v1/login"
AQUALINK_DEVICES_URL = "https://r-api.iaqualink.net/devices.json"

KEEPALIVE_EXPIRY = 30
MIN_SECS_TO_REFRESH = 5
